/*
** Module   : DatabaseTriggerManager.java
** Abstract : Database trigger manager.
**
** Copyright (c) 2009-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description----------------------
** 001 SVL 20091019   @44152 Created initial version. Supports assign and 
**                           delete schema triggers.
** 002 CA  20100527   @44879 Added support for CREATE and "no record is 
**                           available triggers". The CREATE, DELETE and 
**                           ASSIGN triggers will execute on undo too, but the
**                           passed DMO will not be the buffer proxy; instead,
**                           it will be the actual record instance.
** 003 SVL 20130213          Added (de)registerDatabaseTrigger and disable*Triggers functions.
*/

package com.goldencode.p2j.persist.trigger_old;

import com.goldencode.p2j.persist.*;
import com.goldencode.p2j.util.*;
import com.goldencode.p2j.directory.*;
import java.util.*;
import org.apache.commons.logging.*;


/**
 * Manager which responsible for storing database triggers configuration.
 * This configuration is read from the directory. One instance of trigger
 * manager is initialized per database (if it has any enabled triggers).<br/>
 * Rather that providing direct interface for trigger invocation, it provides
 * delegate objects (specific for each kind of DMO interface) which are
 * responsible for this.
 * <p>
 * When a new trigger needs to be added to the directory, following actions 
 * need to be taken:
 * <ol>
 *    <li>Make sure the node under <code>/database/&lt;db_name&gt;/p2j/schema/
 *        schema_triggers/enable</code> is set to <code>true</code>.
 *    </li>
 *    <li>Add the new triggers under the <code>schema_triggers</code> node
 *    following these templates:
 *       <ul>
 *          <li>CREATE triggers. These triggers will be executed right after 
 *          the a new record was created and, if enabled, right after the  
 *          record creation was undone. To enable the trigger for undo, 
 *          just set the <code>on_undo</code> flag for each DMO interface to 
 *          <cde>true</code>. If this flag is missing, then the trigger will
 *          <b>NOT be executed</b> on rollback. <p> The trigger class must 
 *          implement the {@link CreateTrigger} interface.
 *          <pre>
 * &lt;node class="container" name="<b>[trigger class name]</b>"&gt;
 *    &lt;node class="string" name="type"&gt;
 *       &lt;node-attribute name="value" value="<b>CREATE</b>"/&gt;
 *    &lt;/node&gt;
 *    &lt;node class="container" name="targets"&gt;
 *       &lt;node class="container" name="<b>[DMO interface 1]</b>"&gt;
 *          &lt;node class="boolean" name="<b>on_undo</b>"/&gt;
 *             &lt;node-attribute name="value" value="TRUE"/&gt;
 *          &lt;/node&gt;
 *       &lt;node/&gt;
 *       &lt;node class="container" name="<b>[DMO interface 2]</b>"/&gt;
 *       ...
 *    &lt;/node&gt;
 * &lt;/node&gt;
 *          </pre>
 *          </li>
 *          <li>DELETE triggers. These triggers will be executed right before
 *          the record is deleted and, if enabled, right after the record
 *          deletion was undone. To enable the trigger for undo, just set 
 *          the <code>on_undo</code> flag for each DMO interface to 
 *          <cde>true</code>. If this flag is missing, then the trigger will
 *          <b>NOT be executed</b> on rollback. <p>  The trigger class must 
 *          implement the {@link DeleteTrigger} interface.
 *          <pre>
 * &lt;node class="container" name="<b>[trigger class name]</b>"&gt;
 *    &lt;node class="string" name="type"&gt;
 *       &lt;node-attribute name="value" value="<b>DELETE</b>"/&gt;
 *    &lt;/node&gt;
 *    &lt;node class="container" name="targets"&gt;
 *       &lt;node class="container" name="<b>[DMO interface 1]</b>"&gt;
 *          &lt;node class="boolean" name="<b>on_undo</b>"/&gt;
 *             &lt;node-attribute name="value" value="TRUE"/&gt;
 *          &lt;/node&gt;
 *       &lt;node/&gt;
 *       &lt;node class="container" name="<b>[DMO interface 2]</b>"/&gt;
 *       ...
 *    &lt;/node&gt;
 * &lt;/node&gt;
 *          </pre>
 *          </li>
 *          <li>ASSIGN triggers. These triggers will be executed right after
 *          the record's property was changed or, if enabled, right after that
 *          change was undone. To enable the trigger for undo, just set 
 *          the <code>on_undo</code> flag for each DMO interface to 
 *          <cde>true</code>. If this flag is missing, then the trigger will
 *          <b>NOT be executed</b> on rollback. Also, for each DMO interface,
 *          a set of properties must be specified, on which the trigger will
 *          listen for changes.<p>  The trigger class must  implement the 
 *          {@link AssignTrigger} interface.
 *          <pre>
 * &lt;node class="container" name="<b>[trigger class name]</b>"&gt;
 *    &lt;node class="string" name="type"&gt;
 *       &lt;node-attribute name="value" value="<b>ASSIGN</b>"/&gt;
 *    &lt;/node&gt;
 *    &lt;node class="container" name="targets"&gt;
 *       &lt;node class="container" name="<b>[DMO interface 1]</b>"&gt;
 *          &lt;node class="boolean" name="<b>on_undo</b>"/&gt;
 *             &lt;node-attribute name="value" value="TRUE"/&gt;
 *          &lt;/node&gt;
 *          &lt;node class="strings" name="properties"&gt;
 *             &lt;node-attribute name="values" value="<b>[prop1]</b>"/&gt;
 *             &lt;node-attribute name="values" value="<b>[prop2]</b>"/&gt;
 *             ...
 *          &lt;/node&gt;
 *       &lt;node/&gt;
 *       &lt;node class="container" name="<b>[DMO interface 2]</b>"&gt;
 *          &lt;node class="strings" name="properties"&gt;
 *             &lt;node-attribute name="values" value="<b>[prop1]</b>"/&gt;
 *             &lt;node-attribute name="values" value="<b>[prop2]</b>"/&gt;
 *             ...
 *          &lt;/node&gt;
 *       &lt;node/&gt;
 *       ...
 *    &lt;/node&gt;
 * &lt;/node&gt;
 *          </pre>
 *          </li>
 *          <li>"no record is available" triggers. These triggers will execute
 *              only when a "no record is available" error is generated (i.e.
 *              the buffer doesn't reference any record, but a property access
 *              is performed). If the <code>on_undo</code> flag is set, it 
 *              will be ignored - the trigger doesn't need rollback awareness.
 *              <p> The trigger class must implement the 
 *              {@link NoRecordAvailableTrigger} interface.
 *          <pre>
 * &lt;node class="container" name="<b>[trigger class name]</b>"&gt;
 *    &lt;node class="string" name="type"&gt;
 *       &lt;node-attribute name="value" value="<b>NO_RECORD_AVAILABLE</b>"/&gt;
 *    &lt;/node&gt;
 *    &lt;node class="container" name="targets"&gt;
 *       &lt;node class="container" name="<b>[DMO interface 1]</b>"&/gt;
 *       &lt;node class="container" name="<b>[DMO interface 2]</b>"/&gt;
 *       ...
 *    &lt;/node&gt;
 * &lt;/node&gt;
 *          </pre>
 *          </li>
 *       </ul>
 *    </li>
 * </ol>
 * When the <code>CREATE</code>, <code>DELETE</code> or <code>ASSIGN</code>
 * triggers need to be executed on rollback, the trigger class must define an
 * explicit public default constructor (with no parameters) and also a public 
 * c'tor which takes only a <code>boolean</code> parameter. The later one will 
 * be used to instantiate the trigger on rollback.
 */
public class DatabaseTriggerManager
{
   /** Logger */
   private static final Log LOG =
                              LogFactory.getLog(DatabaseTriggerManager.class);

   /** Map of the DMO interfaces to cached delegates for these interfaces. */
   private final Map<Class, DatabaseTriggerDelegate> delegateCache =
                                new HashMap<Class, DatabaseTriggerDelegate>();

   /**
    * Map of the DMO interfaces to lists of classes representing create
    * triggers for these DMO interfaces.
    */
   private final Map<Class, Set<DatabaseTriggerDef>> createTriggers =
                                new HashMap<Class, Set<DatabaseTriggerDef>>();

   /**
    * Map of the DMO interfaces to lists of classes representing delete
    * triggers for these DMO interfaces.
    */
   private final Map<Class, Set<DatabaseTriggerDef>> deleteTriggers =
                                new HashMap<Class, Set<DatabaseTriggerDef>>();

   /**
    * Map of the DMO interfaces to submaps of DMO property names to lists of
    * classes representing assign triggers for these properties.
    */
   private final Map<Class, Map<String, Set<DatabaseTriggerDef>>> 
      assignTriggers = new HashMap<Class, 
                                   Map<String, Set<DatabaseTriggerDef>>>();

   /**
    * Map of the DMO interfaces to lists of classes representing "no record is
    * available" triggers for these DMO interfaces.
    */
   private final Map<Class, Set<DatabaseTriggerDef>> noRecordAvailTriggers =
                                new HashMap<Class, Set<DatabaseTriggerDef>>();

   /**
    * Creates a trigger manager for the specific database.
    *
    * @param database
    *        Database to be served by this trigger manager.
    */
   public DatabaseTriggerManager(Database database)
   {
      // Read schema triggers configuration.
      String basePath = String.format("database/%s/p2j/schema_triggers",
                                      database.getName());

      basePath = Utils.findDirectoryNodePath(null,
                                             basePath,
                                             "container",
                                             false);

      if (basePath == null)
         throw new IllegalStateException("Cannot find database trigger " +
                                         "configuration!");

      DirectoryService ds = DirectoryService.getInstance();

      // enumerate all triggers
      String[] triggerNodes = ds.enumerateNodes(basePath);
      if (triggerNodes != null)
      {
         basePath = new StringBuilder(basePath).append("/%s").toString();

         for (String triggerNode : triggerNodes)
         {
            String triggerNodeId    = String.format(basePath, triggerNode);
            String triggerNodeClass = ds.getNodeClass(triggerNodeId);
            if (!"/meta/class/container".equals(triggerNodeClass))
            {
               continue;
            }

            // validate specified trigger class
            Class triggerClass;
            try
            {
               triggerClass = Class.forName(triggerNode);
            }
            catch(ClassNotFoundException e)
            {
               throw new IllegalStateException(
                              "Cannot load trigger class: " + triggerNode);
            }

            // validate trigger type
            StringBuilder sb = new StringBuilder(triggerNodeId);
            sb.append("/type");

            String typeString = ds.getNodeString(sb.toString(), "value");
            TriggerType type;
            try
            {
               type = TriggerType.valueOf(typeString.toUpperCase());
            }
            catch (IllegalArgumentException e)
            {
               throw new IllegalStateException("Unrecognized trigger type: " +
                                               typeString);
            }

            // get DMO interfaces which are served by this trigger
            sb = new StringBuilder(triggerNodeId);
            sb.append("/targets");

            String targetsNodeId     = sb.toString();
            String[] interfacesNodes = ds.enumerateNodes(targetsNodeId);

            for (String interfaceNode : interfacesNodes)
            {
               sb = new StringBuilder(targetsNodeId);
               sb.append("/");
               sb.append(interfaceNode);

               String interfaceNodeId    = sb.toString();
               String interfaceNodeClass = ds.getNodeClass(interfaceNodeId);
               if (!"/meta/class/container".equals(interfaceNodeClass))
               {
                  continue;
               }

               // validate specified DMO interface
               Class dmoInterface;
               try
               {
                  dmoInterface = Class.forName(interfaceNode);
               }
               catch (ClassNotFoundException e)
               {
                  throw new IllegalStateException(
                                   "Cannot load DMO class: " + interfaceNode);
               }

               // check if the trigger should execute on UNDO too
               sb = new StringBuilder(interfaceNodeId);
               sb.append("/on_undo");
               
               Boolean onUndoVal = ds.getNodeBoolean(sb.toString(), "value");
               boolean onUndo = onUndoVal == null ? false : onUndoVal;
               
               DatabaseTriggerDef trigger = 
                  new DatabaseTriggerDef(triggerClass, onUndo);
               
               // get properties of the DMO interface which are served by this
               // trigger
               sb = new StringBuilder(interfaceNodeId);
               sb.append("/properties");

               String[] properties = ds.getNodeStrings(sb.toString(),
                                                       "values");

               if (type.equals(TriggerType.ASSIGN))
               {
                  // check trigger definition
                  checkTriggerDefinition(trigger, AssignTrigger.class);
                  
                  // we have an assign trigger
                  if (properties == null || properties.length == 0)
                  {
                     throw new IllegalStateException("At least one DMO " +
                       "property should be specified for an assign trigger.");
                  }

                  Set<String> allDMOProperties =
                      PropertyHelper.gettersByProperty(dmoInterface).keySet();

                  // get submap for this DMO interface
                  Map<String, Set<DatabaseTriggerDef>> triggersByProperty =
                                             assignTriggers.get(dmoInterface);
                  if (triggersByProperty == null)
                  {
                     triggersByProperty = 
                        new HashMap<String, Set<DatabaseTriggerDef>>();
                     assignTriggers.put(dmoInterface, triggersByProperty);
                  }

                  for (String property : properties)
                  {
                     // validate DMO property
                     if (!allDMOProperties.contains(property))
                     {
                        sb = new StringBuilder();
                        sb.append("There is no property '");
                        sb.append(property);
                        sb.append("' into the DMO ");
                        sb.append(dmoInterface);

                        throw new IllegalStateException(sb.toString());
                     }

                     // add property -> trigger mapping
                     Set<DatabaseTriggerDef> triggers = 
                        triggersByProperty.get(property);
                     if (triggers == null)
                     {
                        triggers = new HashSet<DatabaseTriggerDef>(1);
                        triggersByProperty.put(property, triggers);
                     }

                     triggers.add(trigger);
                  }
               }
               else
               {
                  Map<Class, Set<DatabaseTriggerDef>> triggerMap = null;
                  Class triggerInterface = null;
                  
                  if (type.equals(TriggerType.DELETE))
                  {
                     // we have a delete trigger
                     triggerMap = deleteTriggers;
                     triggerInterface = DeleteTrigger.class;
                  }
                  else if (type.equals(TriggerType.CREATE))
                  {
                     // we have a create trigger
                     triggerMap = createTriggers;
                     triggerInterface = CreateTrigger.class;
                  }
                  else if (type.equals(TriggerType.NO_RECORD_AVAILABLE))
                  {
                     // we have a "no record is available" trigger
                     triggerMap = noRecordAvailTriggers;
                     triggerInterface = NoRecordAvailableTrigger.class;
                     
                     // reset the onUndo flag
                     onUndo = false;
                  }

                  // check trigger definition
                  checkTriggerDefinition(trigger, triggerInterface);
                  
                  if (triggerMap != null)
                  {
                     // if the trigger type is known
                     Set<DatabaseTriggerDef> triggers = 
                        triggerMap.get(dmoInterface);
                     if (triggers == null)
                     {
                        triggers = new HashSet<DatabaseTriggerDef>(1);
                        triggerMap.put(dmoInterface, triggers);
                     }
   
                     triggers.add(trigger);
                  }
               }
            }
         }
      }

      LOG.info("Database trigger manager initialized for " + database);
   }

   /**
    * Get the delegate object which can invoke database triggers for the
    * specific DMO interface.
    *
    * @param  dmoInterface
    *         DMO interface which will be served by the returned delegate.  
    *
    * @return Delegate object for invocation of the database triggers for the
    *         specific DMO interface or <code>null</code> if there are no
    *         enabled triggers for this interface.
    */
   public DatabaseTriggerDelegate getDelegateForDMO(Class dmoInterface)
   {
      synchronized(delegateCache)
      {
         DatabaseTriggerDelegate delegate = delegateCache.get(dmoInterface);
         if (delegate == null && !delegateCache.containsKey(dmoInterface))
         {
            Set<DatabaseTriggerDef> createBufferTriggers =
                                          createTriggers.get(dmoInterface);
            Set<DatabaseTriggerDef> deleteBufferTriggers =
                                          deleteTriggers.get(dmoInterface);
            Map<String, Set<DatabaseTriggerDef>> assignBufferTriggers =
                                          assignTriggers.get(dmoInterface);
            Set<DatabaseTriggerDef> noRecAvailBufTriggers =
                                      noRecordAvailTriggers.get(dmoInterface);

            if (createBufferTriggers  != null ||
                deleteBufferTriggers  != null ||
                assignBufferTriggers  != null || 
                noRecAvailBufTriggers != null)
            {
               delegate = new DatabaseTriggerDelegate(createBufferTriggers,
                                                      deleteBufferTriggers,
                                                      assignBufferTriggers,
                                                      noRecAvailBufTriggers);
            }
            
            // we put null value into the map if the given DMO interface has
            // no associated triggers 
            delegateCache.put(dmoInterface, delegate);
         }

         return delegate;
      }
   }

   /**
    * Check if the trigger class inherits the expected interface. If the 
    * trigger needs to be executed on undo, check if the trigger class has the
    * following constructors:
    * <ul>
    *    <li>an explicit default (with no parameters) c'tor</li>
    *    <li>a c'tor with only a <code>boolean</code> parameter (i.e. to pass
    *        the <code>onUndo</code> flag.</li>
    * </ul>
    * The default c'tor (with no parameters) will be used to instantiate the
    * trigger for normal operations.
    * 
    * @param  trigger
    *         The trigger definition. 
    * @param  triggerInterface
    *         The interface which must be inherited by the trigger's class.
    *       
    * @throws IllegalStateException
    *         If the trigger's class is not assignable from the expected
    *         interface.
    */
   private void checkTriggerDefinition(DatabaseTriggerDef trigger, 
                                       Class triggerInterface)
   {
      Class clazz = trigger.getTriggerClass();
      if (!triggerInterface.isAssignableFrom(clazz))
      {
         throw new IllegalStateException("The trigger " + 
            clazz.getSimpleName() + " must implement the " + 
            triggerInterface.getSimpleName() + " interface !");
      }
      
      if (trigger.isOnUndo())
      {
         // check if the two c'tors are available
         boolean err = false;
         try
         {
            // check the explicit default c'tor.
            clazz.getConstructor();
         } 
         catch (SecurityException e)
         {
            err = true;
         } 
         catch (NoSuchMethodException e)
         {
            err = true;
         }
         
         if (err)
         {
            throw new IllegalStateException("The trigger " + 
               clazz.getSimpleName() + " must have an explicit default " +
               "c'tor (as it needs to be executed on rollback).");
         }
         
         try
         {
            clazz.getConstructor(boolean.class);
         } 
         catch (SecurityException e)
         {
            err = true;
         } 
         catch (NoSuchMethodException e)
         {
            err = true;
         }
         
         if (err)
         {
            throw new IllegalStateException("The trigger " + 
               clazz.getSimpleName() + " must have a c'tor with a boolean " +
               "parameter (as it needs to be executed on rollback).");
         }
      }
      else
      {
         // check if the default c'tor is available
         boolean err = false;
         try
         {
            clazz.getConstructor();
         } 
         catch (SecurityException e)
         {
            err = true;
         } 
         catch (NoSuchMethodException e)
         {
            err = true;
         }
         
         if (err)
         {
            throw new IllegalStateException("The trigger " + 
             clazz.getSimpleName() + " must have an explicit default c'tor.");
         }
      }
   }

   /**
    * Register a session database trigger within the current scope.
    *
    * @param    event
    *           Database event type.
    * @param    bufferCls
    *           The class of the buffer to which the event is applied.
    * @param    triggerCls
    *           The class of the trigger.
    * @param    container
    *           The containing object (all triggers are implemented as
    *           non-static inner classes, so they must have a reference to
    *           the containing class).
    */
   public static void registerDatabaseTrigger(DatabaseEventType event,
                                              Class<? extends Buffer> bufferCls,
                                              Class<? extends DatabaseTrigger> triggerCls,
                                              Object container)
   {
      registerDatabaseTrigger(event, bufferCls, triggerCls, container, false, false);
   }

   /**
    * Register a session database trigger within the current scope.
    *
    * @param    event
    *           Database event type.
    * @param    bufferCls
    *           The class of the buffer to which the event is applied.
    * @param    triggerCls
    *           The class of the trigger.
    * @param    container
    *           The containing object (all triggers are implemented as
    *           non-static inner classes, so they must have a reference to
    *           the containing class).
    * @param    override
    *           <code>true</code> if OVERRIDE keyword is specified.
    */
   public static void registerDatabaseTrigger(DatabaseEventType event,
                                              Class<? extends Buffer> bufferCls,
                                              Class<? extends DatabaseTrigger> triggerCls,
                                              Object container,
                                              boolean override)
   {
      registerDatabaseTrigger(event, bufferCls, triggerCls, container, override, false);
   }

   /**
    * Register a session database trigger within the current scope.
    *
    * @param    event
    *           Database event type.
    * @param    bufferCls
    *           The class of the buffer to which the event is applied.
    * @param    triggerCls
    *           The class of the trigger.
    * @param    container
    *           The containing object (all triggers are implemented as
    *           non-static inner classes, so they must have a reference to
    *           the containing class).
    * @param    override
    *           <code>true</code> if OVERRIDE keyword is specified.
    * @param    hasOldBuffer
    *           <code>true</code> if the OLD buffer is specified (for a WRITE trigger).
    */
   public static void registerDatabaseTrigger(DatabaseEventType event,
                                              Class<? extends Buffer> bufferCls,
                                              Class<? extends DatabaseTrigger> triggerCls,
                                              Object container,
                                              boolean override,
                                              boolean hasOldBuffer)
   {
      //TODO implement
   }

   /**
    * Deregister a session database trigger for the given event type and buffer class.
    *
    * @param    event
    *           Database event type.
    * @param    bufferCls
    *           The class of the buffer to which the event is applied.
    */
   public static void deregisterDatabaseTrigger(DatabaseEventType event,
                                                Class<? extends Buffer> bufferCls)
   {
      //TODO implement
   }

   /**
    * Disable the triggers associated with the FIND event for the given buffer class.
    *
    * @param bufferCls
    *        The class of the buffer for which the triggers are disabled.
    */
   public static void disableDumpTriggers(Class<? extends Buffer> bufferCls)
   {
      //TODO implement
   }

   /**
    * Disable the triggers associated with the CREATE, WRITE, DELETE and ASSIGN events for the
    * given buffer class.
    *
    * @param bufferCls
    *        The class of the buffer for which the triggers are disabled.
    */
   public static void disableLoadTriggers(Class<? extends Buffer> bufferCls)
   {
      disableLoadTriggers(bufferCls, false);
   }

   /**
    * Disable the triggers associated with the CREATE, WRITE, DELETE and ASSIGN events for the
    * given buffer class.
    *
    * @param bufferCls
    *        The class of the buffer for which the triggers are disabled.
    * @param allowReplication
    *        If <code>true</code> then only CREATE, DELETE, ASSIGN and WRITE triggers are
    *        disabled, and not REPLICATION-CREATE, REPLICATION-DELETE and REPLICATION-WRITE
    *        triggers.
    */
   public static void disableLoadTriggers(Class<? extends Buffer> bufferCls,
                                          boolean allowReplication)
   {
      //TODO implement
   }

   /** Enumeration which defines possible types of database triggers. */
   private enum TriggerType
   {
      /** Corresponds assign triggers. */
      ASSIGN,

      /** Corresponds delete triggers. */
      DELETE,

      /** Corresponds create triggers. */
      CREATE,

      /** Corresponds "no record is available" triggers. */
      NO_RECORD_AVAILABLE
   }
}
